import re
import tensorflow as tf
from os.path import join as path_join
from detext.utils import executor_utils


def create_optimizer(hparams, loss):
    """
    Creates an optimizer training op.
    If the parameter lr_bert is specified, then use another adam for this learning rate.
    """
    tvars = tf.trainable_variables()

    if hparams.use_horovod:
        import horovod.tensorflow as hvd

    # Log trainable variables (with local mode, or chief for ps strategy or rank 0 for hvd training)
    task_type = executor_utils.get_executor_task_type()
    if (hparams.use_horovod is False and task_type in [executor_utils.CHIEF, executor_utils.LOCAL_MODE]) or \
            (hparams.use_horovod is True and hvd.rank() == 0):
        with tf.gfile.Open(path_join(hparams.out_dir, 'network_structure.txt'), 'w') as fout:
            fout.write("# Trainable variables\n")
            total_deep_params = 0
            total_params = 0
            for param in tvars:
                psize = 1
                for s in param.get_shape():
                    psize *= s
                total_params += psize
                if param.name.startswith(hparams.ftr_ext):
                    total_deep_params += psize
                fout.write("  %s, %s, %s\n" % (param.name, str(param.get_shape()), param.op.device))
            fout.write('\n')
            fout.write('# Total trainable params: {}\n'.format(total_params))
            fout.write('# Out of the total trainable params, the total {} parameters: {}\n'
                       .format(hparams.ftr_ext, total_deep_params))

    # Define optimizer parameters
    init_lr = hparams.learning_rate
    num_train_steps = hparams.num_train_steps
    num_warmup_steps = hparams.num_warmup_steps
    lr_bert = hparams.lr_bert

    global_step = tf.train.get_or_create_global_step()

    learning_rate = tf.constant(value=init_lr, shape=[], dtype=tf.float32)

    if hparams.optimizer.startswith("bert_"):
        # Using optimizer with bert's implementation
        # Implements linear decay of the learning rate.
        learning_rate = tf.train.polynomial_decay(
            learning_rate,
            global_step,
            num_train_steps,
            end_learning_rate=0.0,
            power=1.0,
            cycle=False)

        # Implements linear warmup. I.e., if global_step < num_warmup_steps, the
        # learning rate will be `global_step/num_warmup_steps * init_lr`.
        if num_warmup_steps:
            global_steps_int = tf.cast(global_step, tf.int32)
            warmup_steps_int = tf.constant(num_warmup_steps, dtype=tf.int32)

            global_steps_float = tf.cast(global_steps_int, tf.float32)
            warmup_steps_float = tf.cast(warmup_steps_int, tf.float32)

            warmup_percent_done = global_steps_float / warmup_steps_float
            warmup_learning_rate = init_lr * warmup_percent_done

            is_warmup = tf.cast(global_steps_int < warmup_steps_int, tf.float32)
            learning_rate = ((1.0 - is_warmup) * learning_rate + is_warmup * warmup_learning_rate)

        name_2_optimizer = {
            'bert_adam': AdamWeightDecayOptimizer,
            'bert_lamb': LAMBOptimizer
        }
        OptimizerFunc = name_2_optimizer[hparams.optimizer]

        # It is recommended that you use this optimizer for fine tuning, since this
        # is how the model was trained (note that the Adam/Lamb m/v variables are NOT
        # loaded from init_checkpoint.)
        optimizer = OptimizerFunc(
            learning_rate=learning_rate,
            weight_decay_rate=0.01,
            beta_1=0.9,
            beta_2=0.999,
            epsilon=1e-6,
            exclude_from_weight_decay=["LayerNorm", "layer_norm", "bias"])

        if hparams.use_horovod:
            # Horovod's distributed optimizer handles allreduce calls, synchronous only
            optimizer = hvd.DistributedOptimizer(optimizer, sparse_as_dense=True)
            grads_and_vars = optimizer.compute_gradients(loss, tvars)
            grads = [grad for grad, var in grads_and_vars]
            tvars = [var for grad, var in grads_and_vars]
        else:
            grads = tf.gradients(loss, tvars)

        grads, grad_norm = tf.clip_by_global_norm(grads, clip_norm=1.0)

        if lr_bert is None:
            # If not a separate learning rate for bert (lr_bert) is specified,
            # all components use the same learning rate
            train_op = optimizer.apply_gradients(
                zip(grads, tvars), global_step=global_step)

            # Normally the global step update is done inside of `apply_gradients`.
            # However, `AdamWeightDecayOptimizer` doesn't do this. But if you use
            # a different optimizer, you should probably take this line out.
            new_global_step = global_step + 1
            train_op = tf.group(train_op, [global_step.assign(new_global_step)])
        else:
            # the BERT components will use another learning rate
            optimizer_bert = OptimizerFunc(
                learning_rate=learning_rate * lr_bert / init_lr,
                weight_decay_rate=0.01,
                beta_1=0.9,
                beta_2=0.999,
                epsilon=1e-6,
                exclude_from_weight_decay=["LayerNorm", "layer_norm", "bias"])
            if hparams.use_horovod:
                # Treat the bert optimizer the same as the original optimizer: wrapped with horovod
                optimizer_bert = hvd.DistributedOptimizer(optimizer_bert, sparse_as_dense=True)

            bert_grad, bert_tvars = [], []
            other_grad, other_tvars = [], []
            for grad, tvar in zip(grads, tvars):
                if tvar is not None and grad is not None:
                    if tvar.name.startswith('bert'):
                        bert_grad.append(grad)
                        bert_tvars.append(tvar)
                        print('****bert param:', tvar.name)
                    else:
                        other_grad.append(grad)
                        other_tvars.append(tvar)
                        print('****other param:', tvar.name)
            print('--------------\n', '# of bert', len(bert_grad), '# of other', len(other_grad), '\n--------------')
            bert_train_op = optimizer_bert.apply_gradients(
                zip(bert_grad, bert_tvars), global_step=global_step)
            other_train_op = optimizer.apply_gradients(
                zip(other_grad, other_tvars), global_step=global_step)

            new_global_step = global_step + 1
            train_op = tf.group(bert_train_op, other_train_op, [global_step.assign(new_global_step)])

        return train_op, grad_norm, learning_rate

    elif hparams.optimizer == "sgd":
        opt = tf.train.GradientDescentOptimizer(learning_rate)
    elif hparams.optimizer == "adam":
        opt = tf.train.AdamOptimizer(learning_rate)
    else:
        raise ValueError("Only support sgd/adam/bert_adam as optimizer option")

    # Gradients
    gradients = tf.gradients(loss, tvars, colocate_gradients_with_ops=True)
    clipped_gradients, grad_norm = tf.clip_by_global_norm(gradients, hparams.max_gradient_norm)
    train_op = opt.apply_gradients(zip(clipped_gradients, tvars), global_step=global_step)

    return train_op, grad_norm, learning_rate


class AdamWeightDecayOptimizer(tf.train.Optimizer):
    """A basic Adam optimizer that includes "correct" L2 weight decay."""

    def __init__(self,
                 learning_rate,
                 weight_decay_rate=0.0,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-6,
                 exclude_from_weight_decay=None,
                 name="AdamWeightDecayOptimizer"):
        """Constructs a AdamWeightDecayOptimizer."""
        super(AdamWeightDecayOptimizer, self).__init__(False, name)

        self.learning_rate = learning_rate
        self.weight_decay_rate = weight_decay_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.exclude_from_weight_decay = exclude_from_weight_decay

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        """See base class."""
        assignments = []
        for (grad, param) in grads_and_vars:
            if grad is None or param is None:
                continue

            param_name = self._get_variable_name(param.name)

            m = tf.get_variable(
                name=param_name + "/adam_m",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())
            v = tf.get_variable(
                name=param_name + "/adam_v",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())

            # Standard Adam update.
            next_m = (tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad))
            next_v = (tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2, tf.square(grad)))

            update = next_m / (tf.sqrt(next_v) + self.epsilon)

            # Just adding the square of the weights to the loss function is *not*
            # the correct way of using L2 regularization/weight decay with Adam,
            # since that will interact with the m and v parameters in strange ways.
            #
            # Instead we want ot decay the weights in a manner that doesn't interact
            # with the m/v parameters. This is equivalent to adding the square
            # of the weights to the loss with plain (non-momentum) SGD.
            if self._do_use_weight_decay(param_name):
                update += self.weight_decay_rate * param

            update_with_lr = self.learning_rate * update

            next_param = param - update_with_lr

            assignments.extend(
                [param.assign(next_param),
                 m.assign(next_m),
                 v.assign(next_v)])
        return tf.group(*assignments, name=name)

    def _do_use_weight_decay(self, param_name):
        """Whether to use L2 weight decay for `param_name`."""
        if not self.weight_decay_rate:
            return False
        if self.exclude_from_weight_decay:
            for r in self.exclude_from_weight_decay:
                if re.search(r, param_name) is not None:
                    return False
        return True

    def _get_variable_name(self, param_name):
        """Get the variable name from the tensor name."""
        m = re.match("^(.*):\\d+$", param_name)
        if m is not None:
            param_name = m.group(1)
        return param_name


class LAMBOptimizer(tf.train.Optimizer):
    """
    Optimizer that implements the Layer-wise Adaptive Moments (LAMB).
    See paper [Large Batch Optimization for Deep Learning: Training BERT
        in 76 minutes](https://arxiv.org/abs/1904.00962).
    Official implementation from tensorflow addon tensorflow_addons/optimizers/lamb.py

    Note that we does not apply adam bias correction for the moments estimates to keep it similar to the original AdamW
    optimizer in BERT pretraining (which is the AdamWeightDecayOptimizer here). The only difference for the
    LAMBOptimizer compared to AdamWeightDecayOptimizer implementation is the added layer adaptation.
    """

    def __init__(self,
                 learning_rate,
                 weight_decay_rate=0.01,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-6,
                 exclude_from_weight_decay=None,
                 name="LAMBOptimizer"):
        """Constructs a LAMBOptimizer."""
        super(LAMBOptimizer, self).__init__(False, name)

        self.learning_rate = learning_rate
        self.weight_decay_rate = weight_decay_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.exclude_from_weight_decay = exclude_from_weight_decay

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        """See base class."""
        assignments = []
        for (grad, param) in grads_and_vars:
            if grad is None or param is None:
                continue

            param_name = self._get_variable_name(param.name)

            m = tf.get_variable(
                name=param_name + "/lamb_m",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())
            v = tf.get_variable(
                name=param_name + "/lamb_v",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())

            # Standard Adam update.
            next_m = (tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad))
            next_v = (tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2, tf.square(grad)))

            update = next_m / (tf.sqrt(next_v) + self.epsilon)

            # Just adding the square of the weights to the loss function is *not*
            # the correct way of using L2 regularization/weight decay with Adam,
            # since that will interact with the m and v parameters in strange ways.
            #
            # Instead we want ot decay the weights in a manner that doesn't interact
            # with the m/v parameters. This is equivalent to adding the square
            # of the weights to the loss with plain (non-momentum) SGD.
            if self._do_use_weight_decay(param_name):
                update += self.weight_decay_rate * param

            # lamb layer adaptation
            w_norm = tf.norm(param, ord=2)
            g_norm = tf.norm(update, ord=2)

            ratio = tf.where(
                tf.greater(w_norm, 0),
                tf.where(tf.greater(g_norm, 0), (w_norm / g_norm), 1.0), 1.0)

            update_with_lr = self.learning_rate * ratio * update

            next_param = param - update_with_lr

            assignments.extend(
                [param.assign(next_param),
                 m.assign(next_m),
                 v.assign(next_v)])
        return tf.group(*assignments, name=name)

    def _do_use_weight_decay(self, param_name):
        """Whether to use L2 weight decay for `param_name`."""
        if not self.weight_decay_rate:
            return False
        if self.exclude_from_weight_decay:
            for r in self.exclude_from_weight_decay:
                if re.search(r, param_name) is not None:
                    return False
        return True

    def _get_variable_name(self, param_name):
        """Get the variable name from the tensor name."""
        m = re.match("^(.*):\\d+$", param_name)
        if m is not None:
            param_name = m.group(1)
        return param_name



"""
lambda rank implementation.
"""
import tensorflow as tf

from detext.train import metrics


class LambdaRank(tf.layers.Layer):
    """
    LambdaRank as the learning-to-rank function.
    For now, the current version does not implement the "lambda" part.
    """

    def __init__(self, lambda_metric=None):
        self.lambda_metric = lambda_metric
        super(LambdaRank, self).__init__()

    def build(self, _):
        super(LambdaRank, self).build(_)

    def call(self, scores, labels, group_size):
        """ Compute the pairwise loss.

        :param scores: A tensor with shape [batch_size, max_group_size].  For each batch, the first element is the score of
        correct answer.
        :param labels: A matrix with shape [batch_size, max_group_size].  The true scores of each document.
        :param group_size: A vector with shape [batch_size], indicating how many documents each query has.
        :return: lambdarank loss and mask. Each with shape [batch_size, max_group_size, max_group_size]
        """
        # for a query, compute the pairwise doc score diff of any two documents.
        pair_score_diff = tf.expand_dims(scores, axis=2) - tf.expand_dims(scores, axis=1)
        # compute the loss
        loss = -1 * tf.log_sigmoid(pair_score_diff)
        # now loss is a [batch_size, max_group_size, max_group_size] tensor that contains all pairwise loss.
        # we only need to keep a subset of the pairs.
        # the first mask is from group_size
        group_size_mask = tf.sequence_mask(group_size, maxlen=tf.shape(scores)[1], dtype=tf.float32)
        group_size_mask = tf.expand_dims(group_size_mask, axis=2) * tf.expand_dims(group_size_mask, axis=1)
        # the second mask is from label; only keep the pairs that 1st label value is larger than 2nd value
        label_mask = tf.expand_dims(labels, axis=2) - tf.expand_dims(labels, axis=1)
        label_mask = tf.cast(tf.greater(label_mask, tf.zeros_like(label_mask)), dtype=tf.float32)
        pairwise_mask = group_size_mask * label_mask
        loss *= pairwise_mask

        if self.lambda_metric:
            # compute each element's rank
            rank_mat = metrics.compute_rank(scores, group_size)
            if self.lambda_metric['metric'] == 'ndcg':
                # ideal dcg
                idcg = metrics.compute_dcg(labels, labels, group_size, self.lambda_metric['topk'])
                # delta_score
                delta_score = tf.expand_dims(labels, axis=2) - tf.expand_dims(labels, axis=1)
                # delta rank
                reci_log_rank = tf.log(2.0) / tf.log(tf.cast(rank_mat, dtype=tf.float32) + 1)
                delta_rank = tf.expand_dims(reci_log_rank, axis=2) - tf.expand_dims(reci_log_rank, axis=1)
                # delta_ndcg = |delta_score * delta_rank| / idcg
                delta_ndcg = tf.abs(delta_score * delta_rank) / tf.expand_dims(tf.expand_dims(idcg, 1), 1)
                # lambda loss
                loss *= delta_ndcg

        return loss, pairwise_mask



import tensorflow as tf

from detext.utils.model_utils import init_word_embedding


class IdEmbedModel:
    """Model that produces id embedding for given doc/usr id fields

    Attributes:
        doc_ftrs  Tensor Shape=[batch_size, max_group_size, num_doc_id_fields, num_units_for_id_ftr]
        usr_ftrs  Tensor Shape=[batch_size, num_usr_id_fields, num_units_for_id_ftr]
    """

    def __init__(self,
                 doc_id_fields,
                 usr_id_fields,
                 hparams,
                 mode
                 ):
        """
        :param doc_id_fields: list(Tensor)  List of document fields. Each has
            shape=[batch_size, max_group_size, num_doc_id]
        :param usr_id_fields: list(Tensor)  List of user fields. Each has
            shape=[batch_size, num_usr_id]
        :param hparams: HParams  Hyper parameters
        :param mode: TRAIN/EVAL/INFER
        """
        self._usr_id_fields = usr_id_fields
        self._doc_id_fields = doc_id_fields
        self._hparams = hparams
        self._mode = mode
        self.id_ftr_size = hparams.num_units_for_id_ftr

        with tf.variable_scope("id_embed", dtype=tf.float32):
            if usr_id_fields is not None or usr_id_fields is not None:
                self.embedding = init_word_embedding(
                    {"we_trainable": hparams.we_trainable_for_id_ftr,
                     "we_file": hparams.we_file_for_id_ftr,
                     "vocab_size": hparams.vocab_size_for_id_ftr,
                     "num_units": hparams.num_units_for_id_ftr
                     }, self._mode)
            self.usr_ftrs = self.apply_embed_on_usr_id() if usr_id_fields is not None else None
            self.doc_ftrs = self.apply_embed_on_doc_id() if doc_id_fields is not None else None

    def apply_embed_on_usr_id(self):
        """Applies embedding lookup and averaging for user id features

        :return Tensor Shape=[batch_size, num_usr_id_fields, num_units_for_id_ftr]
        """
        hparams = self._hparams
        usr_ftrs = []
        for i, usr_field in enumerate(self._usr_id_fields):
            seq_mask = tf.cast(tf.not_equal(usr_field, hparams.pad_id_for_id_ftr),
                               dtype=tf.float32)  # [batch_size, num_usr_id]
            seq_mask = tf.expand_dims(seq_mask, axis=-1)  # [batch_size, num_usr_id, 1]
            seq_length = tf.reduce_sum(seq_mask, axis=-2)  # [batch_size, 1]

            usr_id_embeddings = tf.nn.embedding_lookup(
                self.embedding, usr_field)  # [batch_size, num_usr_id, num_units_for_id_ftr]
            sum_usr_id_embedding = tf.reduce_sum(
                usr_id_embeddings * seq_mask, axis=1)  # [batch_size, num_units_for_id_ftr]
            usr_id_avg_embedding = tf.div_no_nan(sum_usr_id_embedding, seq_length)  # [batch_size, num_units_for_id_ftr]
            usr_ftrs.append(usr_id_avg_embedding)
        return tf.stack(usr_ftrs, axis=1)

    def apply_embed_on_doc_id(self):
        """Applies embedding lookup and averaging for doc id features

        :return Tensor Shape=[batch_size, max_group_size, num_doc_id_fields, num_units_for_id_ftr]
        """
        hparams = self._hparams

        doc_ftrs = []
        for i, doc_field in enumerate(self._doc_id_fields):
            seq_mask = tf.cast(tf.not_equal(doc_field, hparams.pad_id_for_id_ftr),
                               dtype=tf.float32)  # [batch_size, max_group_size, num_doc_id]
            seq_mask = tf.expand_dims(seq_mask, axis=-1)  # [batch_size, max_group_size, num_doc_id, 1]
            seq_length = tf.reduce_sum(seq_mask, axis=-2,)  # [batch_size, max_group_size, 1]

            doc_id_embeddings = tf.nn.embedding_lookup(
                self.embedding, doc_field, )  # [batch_size, max_group_size, num_doc_id, num_units_for_id_ftr]
            sum_doc_id_embedding = tf.reduce_sum(
                doc_id_embeddings * seq_mask, axis=2)  # [batch_size, max_group_size, num_units_for_id_ftr]
            doc_id_avg_embedding = tf.div_no_nan(sum_doc_id_embedding,
                                                 seq_length)  # [batch_size, max_group_size, num_units_for_id_ftr]
            doc_ftrs.append(doc_id_avg_embedding)
        return tf.stack(doc_ftrs, axis=2)



"""Utility functions related to distributed training executor setup."""
import json
import logging
import os
import tensorflow as tf

CHIEF = 'chief'
EVALUATOR = 'evaluator'
WORKER = 'worker'
PARAMETER_SERVER = 'ps'
LOCAL_MODE = 'local'


def get_executor_task_type():
    """Get executor task type (chief/evaluator/worker/ps or local) from TF_CONFIG."""
    tf.logging.info("Getting executor context info...")
    tf_config = os.environ.get("TF_CONFIG")

    # With multiworker training, tf_config contains tasks in the cluster, and each task's type in the cluster.
    # To get ther worker/evaluator status, need to fetch corresponding fields in the config (json format).
    # Read more at https://www.tensorflow.org/guide/distributed_training#setting_up_tf_config_environment_variable
    if tf_config:
        tf_config_json = json.loads(tf_config)

        # Logging the status of current worker/evaluator
        logging.info("Running with TF_CONFIG: {}".format(tf_config_json))
        task_type = tf_config_json.get('task', {}).get('type')
        logging.info("=========== Current executor task type: {} ==========".format(task_type))
        return task_type
    else:
        logging.info("=========== No TF_CONFIG found. Running {} mode. ==========".format(LOCAL_MODE))
        return LOCAL_MODE



"""
Utility function for vocabulary
"""
import gzip
import os
import re
import six
import string
import tensorflow as tf
from tensorflow.python.ops import lookup_ops


def convert_to_unicode(text):
    """Converts `text` to Unicode (if it's not already), assuming utf-8 input."""
    if six.PY3:
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode("utf-8", "ignore")
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    elif six.PY2:
        if isinstance(text, str):
            return text.decode("utf-8", "ignore")
        # In py2, six.text_type = unicode. We use six.text_type to pass the pyflake checking
        elif isinstance(text, six.text_type):
            return text
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    else:
        raise ValueError("Not running on Python2 or Python 3?")


def convert_to_str(text):
    """
    Converts `text` to str (if it's not already).
    Used when writing text to file with `open(filename, 'w')`.
    """
    if six.PY2:  # In python 2, str is byte
        if isinstance(text, str):
            return text
        else:
            return convert_to_unicode(text).encode('utf-8')
    elif six.PY3:  # In python3, str is unicode
        if isinstance(text, str):
            return text
        else:
            return convert_to_unicode(text)
    raise ValueError("Not running on Python2 or Python 3?")


def convert_to_bytes(text):
    """
    Converts `text` to bytes (if it's not already).
    Used when generating tfrecords. More specifically, in function call `tf.train.BytesList(value=[<bytes1>, <bytes2>, ...])`
    """
    if six.PY2:
        return convert_to_str(text)  # In python2, str is byte
    elif six.PY3:
        if isinstance(text, bytes):
            return text
        else:
            return convert_to_unicode(text).encode('utf-8')
    else:
        raise ValueError("Not running on Python2 or Python 3?")


def strip(s):
    """Strips ascii whitespace characters off string s."""
    return s.strip(string.whitespace)


def split(s):
    """Split string s by whitespace characters."""
    whitespace_lst = [re.escape(ws) for ws in string.whitespace]
    pattern = re.compile('|'.join(whitespace_lst))
    return pattern.split(s)


def read_vocab(input_file):
    """Read vocabulary file and return a dict"""
    if input_file is None:
        return None

    vocab = {}
    if input_file.endswith('.gz'):
        f = tf.gfile.Open(input_file, 'r')
        fin = gzip.GzipFile(fileobj=f)
    else:
        fin = tf.gfile.Open(input_file, 'r')
    for line in fin:
        word = split(strip(line))[0]
        vocab[word] = len(vocab)
    fin.close()
    return vocab


def read_tf_vocab(input_file, UNK):
    """Read vocabulary and return a tf hashtable"""
    if input_file is None:
        return None

    keys, values = [], []
    if input_file.endswith('.gz'):
        f = tf.gfile.Open(input_file, 'r')
        fin = gzip.GzipFile(fileobj=f)
    else:
        fin = tf.gfile.Open(input_file, 'r')
    for line in fin:
        word = split(strip(line))[0]
        keys.append(word)
        values.append(len(values))
    fin.close()
    UNK_ID = keys.index(UNK)
    vocab_table = lookup_ops.HashTable(lookup_ops.KeyValueTensorInitializer(
        tf.constant(keys), tf.constant(values)), UNK_ID)
    return vocab_table


def extract_text_data(input_dir, output_file, text_fields):
    """
    Extract text data from tfrecords.  The data will be used for word embedding pretraining.
    """
    with tf.gfile.Open(output_file, 'w') as fout:
        for file in tf.gfile.ListDirectory(input_dir):
            input_file = os.path.join(input_dir, file)
            print(input_file)
            for example in tf.python_io.tf_record_iterator(input_file):
                result = tf.train.Example.FromString(example)
                for field in text_fields:
                    text_values = result.features.feature[field].bytes_list.value
                    for text in text_values:
                        text = convert_to_unicode(text)
                        text = text.strip()
                        if ' ' not in text:
                            continue
                        fout.write(text + '\n')


if __name__ == '__main__':
    input_dir = '/home/wguo/data/people_v6/NAV/train/'
    output_file = '/home/wguo/data/people_v6/NAV/glove/text'
    text_fields = ['query', 'doc_headlines', 'doc_currCompanies', 'doc_pastCompanies', 'doc_currTitles',
                   'doc_pastTitles', 'doc_currSchools', 'doc_pastSchools']
    extract_text_data(input_dir, output_file, text_fields)
    # max_vocab_size = 100000
    # generate_vocab_file(input_dir, output_file, max_vocab_size)



"""
The model that uses BERT to extract features from texts, and generate a score for a query/doc pair
Different from CNN, there is only one BERT model that generates the text embedding for all fields.
"""

import tensorflow as tf

from detext.bert_embedding_extraction.bert_utils import create_bert_model, init_from_checkpoint


class BertModel(object):
    """Apply BERT to convert text to a fix length embedding."""

    def __init__(self,
                 query,
                 doc_fields,
                 usr_fields,
                 hparams,
                 mode):
        """
        Apply BERT to convert text to features
        :param query: Shape=[batch_size, query_length]
        :param doc_fields: list(Tensor) or a pre-computed Tensor  List of document fields. Each has
            shape=[batch_size, max_group_size, doc_field_length]
        :param usr_fields: list(Tensor) or a pre-computed Tensor  List of user fields. Each has
            shape=[batch_size, usr_field_length]
        :param hparams:
        :param mode:
        """
        self.query = query
        self.doc_fields = doc_fields
        self.usr_fields = usr_fields
        self.hparams = hparams
        self.mode = mode

        self.query_ftrs, self.doc_ftrs, self.usr_ftrs = self.apply_bert()

    def apply_bert(self):
        """Sends the query and doc_fields to BERT to get text embedding.

        If query is None, the returned query_ftrs is also None
        :return query_ftrs,doc_ftrs,usr_ftrs  shape(query_ftrs)=[batch_size, num_units],
            shape(doc_ftrs)=[batch_size, max_group_size, num_doc_fields, num_units],
            shape(usr_ftrs)=[batch_size, num_doc_fields, num_units]
        """
        hparams = self.hparams
        query = self.query
        has_query = query is not None
        has_doc = self.doc_fields is not None

        # at least one of query and doc_fields should be true
        if not (has_query or has_doc):
            raise ValueError('query and doc_fields cannot both be None')

        doc_fields = self.doc_fields
        usr_fields = self.usr_fields

        # if doc is None, we cannot infer batch_size and max_group_size from doc
        batch_size = tf.shape(doc_fields[0])[0] if has_doc else tf.shape(query)[0]
        max_group_size = tf.shape(doc_fields[0])[1] if has_doc else None

        if not has_query:
            # Instead of removing the query sequence in the input id array fed into BERT, we use a fake query with
            #   the same size as the first document field in the first document. The content of this query will NOT
            #   be used anywhere. It is just to save effort on code changes handling the case when `query is None`
            query = tf.ones(dtype=tf.int32, shape=tf.shape(doc_fields[0][:, 0, :]))

        # 1. combine query and doc_fields into one matrix
        # find largest text length value
        max_text_len = tf.shape(query)[-1]
        max_text_len_array = [max_text_len]

        # Add doc field length info
        if type(doc_fields) is list:
            for doc_field in doc_fields:
                doc_field_text_len = tf.shape(doc_field)[-1]
                max_text_len_array.append(doc_field_text_len)
                max_text_len = tf.maximum(max_text_len, doc_field_text_len)

        # Add usr field length info
        if type(usr_fields) is list:
            for usr_field in usr_fields:
                usr_field_text_len = tf.shape(usr_field)[-1]
                max_text_len_array.append(usr_field_text_len)
                max_text_len = tf.maximum(max_text_len, usr_field_text_len)

        # pad data
        padded_query = tf.pad(query, [[0, 0], [0, max_text_len - max_text_len_array[0]]],
                              constant_values=hparams.pad_id)
        bert_input_ids_array = [padded_query]

        if type(doc_fields) is list:
            for doc_field, doc_field_max_text_len in zip(doc_fields, max_text_len_array[1:]):
                # Shape=[batch_size*max_group_size, doc_field_max_text_len]
                doc_field_2d = tf.reshape(doc_field, shape=[-1, doc_field_max_text_len])
                padded_doc_field = tf.pad(
                    doc_field_2d, paddings=[[0, 0], [0, max_text_len - doc_field_max_text_len]],
                    constant_values=hparams.pad_id)
                bert_input_ids_array.append(padded_doc_field)

        if type(usr_fields) is list:
            doc_field_end = 1 + len(doc_fields) if type(doc_fields) is list else 1

            for usr_field, usr_field_max_text_len in zip(usr_fields, max_text_len_array[doc_field_end:]):
                # Shape=[batch_size, usr_field_max_text_len]
                padded_usr_field = tf.pad(usr_field, paddings=[[0, 0], [0, max_text_len - usr_field_max_text_len]],
                                          constant_values=hparams.pad_id)
                bert_input_ids_array.append(padded_usr_field)

        self.bert_input_ids = tf.concat(bert_input_ids_array, axis=0)
        self.bert_input_mask = tf.cast(tf.not_equal(self.bert_input_ids, hparams.pad_id), dtype=tf.int32)

        # 2. apply bert
        self.text_ftr_size = hparams.num_units

        is_training = self.mode == tf.estimator.ModeKeys.TRAIN

        bertm = create_bert_model(
            bert_config_path=hparams.bert_config_file,
            is_training=(is_training and hparams.use_bert_dropout),
            input_ids=self.bert_input_ids,
            input_mask=self.bert_input_mask,
            use_one_hot_embeddings=False
        )

        # initialize bert with pretrained checkpoint
        if is_training and self.hparams.bert_checkpoint is not None:
            init_from_checkpoint(self.hparams.bert_checkpoint)

        # 3. get query embedding and doc_fields embedding
        last_layer = tf.squeeze(bertm.get_all_encoder_layers()[-1][:, 0:1, :], axis=1)
        pooled_output = tf.layers.dense(
            last_layer,
            self.text_ftr_size,
            activation=tf.tanh,
            name='bert/last_layer_cls'
        )  # shape = [# of texts, bert_dim]

        # dropout
        if is_training:
            pooled_output = tf.nn.dropout(pooled_output, keep_prob=0.9)

        # Get query features
        query_ftrs = pooled_output[0:batch_size] if has_query else None

        # Get document features
        if type(doc_fields) is not list:
            # If doc_fields is not a list, then it is a precomputed embedding tensor
            doc_ftrs = self.doc_fields
            doc_field_end = batch_size
        else:
            doc_field_start = batch_size
            doc_ftrs = []
            for _ in range(len(doc_fields)):
                doc_field_end = doc_field_start + batch_size * max_group_size
                dftrs = pooled_output[doc_field_start: doc_field_end]
                dftrs = tf.reshape(dftrs, shape=[batch_size, max_group_size, self.text_ftr_size])
                doc_ftrs.append(dftrs)
                doc_field_start = doc_field_end
            doc_ftrs = tf.stack(doc_ftrs, axis=2)  # shape = [batch_size, max_group_size, num_doc_fields, num_units]

        # Get user features
        if type(usr_fields) is not list:
            usr_ftrs = self.usr_fields
        else:
            usr_field_start = doc_field_end
            usr_ftrs = []
            for _ in range(len(usr_fields)):
                usr_field_end = usr_field_start + batch_size
                uftrs = pooled_output[usr_field_start: usr_field_end]
                usr_ftrs.append(uftrs)
                usr_field_start = usr_field_end
            usr_ftrs = tf.stack(usr_ftrs, axis=1)  # shape=[batch_size, num_usr_fields, num_units]

        return query_ftrs, doc_ftrs, usr_ftrs



import codecs
import json
import os
import pickle
import random

import numpy as np
import tensorflow as tf
from detext.model.bert import modeling
from detext.utils import test_utils
from detext.utils import vocab_utils


def force_set_hparam(hparams, name, value):
    """
    Removes name from hparams and sets hparams.name == value.
    This function is introduced because hparams.set_hparam(name, value) requires value to be of the same type as the
        existing hparam.get(name) if name is already set in hparam
    """
    hparams.del_hparam(name)
    hparams.add_hparam(name, value)


def get_config_proto(log_device_placement=False, allow_soft_placement=True):
    # GPU options:
    # https://www.tensorflow.org/versions/r0.10/how_tos/using_gpu/index.html
    config_proto = tf.ConfigProto(
        log_device_placement=log_device_placement,
        allow_soft_placement=allow_soft_placement)
    config_proto.gpu_options.allow_growth = True
    return config_proto


def save_hparams(out_dir, hparams):
    """Save hparams."""
    hparams_file = os.path.join(out_dir, "hparams")
    print("  saving hparams to %s" % hparams_file)
    with codecs.getwriter("utf-8")(tf.gfile.GFile(hparams_file, "wb")) as f:
        f.write(hparams.to_json())


def load_hparams(model_dir):
    """Load hparams from an existing model directory."""
    hparams_file = os.path.join(model_dir, "hparams")
    if tf.gfile.Exists(hparams_file):
        print("# Loading hparams from %s" % hparams_file)
        with codecs.getreader("utf-8")(tf.gfile.GFile(hparams_file, "rb")) as f:
            try:
                hparams_values = json.load(f)
                hparams = tf.contrib.training.HParams(**hparams_values)
            except ValueError:
                print("  can't load hparams file")
                return None
        return hparams
    else:
        return None


def extend_hparams(hparams):
    # Sanity check for RNN related hparams
    assert hparams.unit_type in ['lstm', 'gru', 'layer_norm_lstm'], 'Only support lstm/gru/layer_norm_lstm as unit_type'
    assert hparams.num_layers > 0, 'num_layers must be larger than 0'
    assert hparams.num_residual_layers >= 0, 'num_residual_layers must >= 0'
    assert 0 <= hparams.forget_bias <= 1, 'forget_bias must be within [0.0, 1.0]'
    assert 0 <= hparams.rnn_dropout <= 1, 'rnn_dropout must be within [0.0, 1.0]'

    # Get number of doc/usr text fields
    num_doc_fields = sum([name.startswith('doc_') for name in hparams.feature_names])
    hparams.add_hparam("num_doc_fields", num_doc_fields)
    num_usr_fields = sum([name.startswith('usr_') for name in hparams.feature_names])
    hparams.add_hparam("num_usr_fields", num_usr_fields)

    # Get number of doc/usr id fields
    num_doc_id_fields = sum([name.startswith('docId_') for name in hparams.feature_names])
    hparams.add_hparam("num_doc_id_fields", num_doc_id_fields)
    num_usr_id_fields = sum([name.startswith('usrId_') for name in hparams.feature_names])
    hparams.add_hparam("num_usr_id_fields", num_usr_id_fields)
    if num_doc_id_fields > 0 or num_usr_id_fields > 0:
        assert hparams.vocab_file_for_id_ftr is not None, \
            "Must provide vocab_field_for_id_ftr arg when id features are provided"

    # find vocab size, pad id from vocab file
    vocab_table = vocab_utils.read_vocab(hparams.vocab_file)
    hparams.add_hparam("vocab_size", len(vocab_table))
    hparams.pad_id = vocab_table[hparams.PAD]

    # find vocab size, pad id from vocab file for id features
    if hparams.vocab_file_for_id_ftr is not None:
        vocab_table_for_id_ftr = vocab_utils.read_vocab(hparams.vocab_file_for_id_ftr)
        hparams.add_hparam("vocab_size_for_id_ftr", len(vocab_table_for_id_ftr))
        hparams.pad_id_for_id_ftr = vocab_table_for_id_ftr[hparams.PAD_FOR_ID_FTR]

    # if there is bert config, check compatibility of between bert parameters and existing parameters
    if hparams.bert_config_file:
        hparams.bert_config = modeling.BertConfig.from_json_file(hparams.bert_config_file)
        assert hparams.bert_config.vocab_size == hparams.vocab_size

    # The regex pattern to add a white space before and after. Used for processing text fields.
    tok2regex_pattern = {'plain': None, 'punct': r'(\pP)'}
    hparams.regex_replace_pattern = tok2regex_pattern[hparams.tokenization]

    # if not using cnn models, then disable cnn parameters
    if hparams.ftr_ext != 'cnn':
        hparams.filter_window_sizes = [0]

    assert hparams.pmetric is not None, "Please set your primary evaluation metric using --pmetric option"
    assert hparams.pmetric != 'confusion_matrix', 'confusion_matrix cannot be used as primary evaluation metric.'

    # Set all relevant evaluation metrics
    all_metrics = hparams.all_metrics if hparams.all_metrics else [hparams.pmetric]
    assert hparams.pmetric in all_metrics, "pmetric must be within all_metrics"
    force_set_hparam(hparams, "all_metrics", all_metrics)

    # lambda rank
    if hparams.lambda_metric is not None and hparams.lambda_metric == 'ndcg':
        setattr(hparams, 'lambda_metric', {'metric': 'ndcg', 'topk': 10})
    else:
        setattr(hparams, 'lambda_metric', None)
    # feature normalization
    if hparams.std_file:
        # read normalization file
        print('read normalization file')
        ftr_mean, ftr_std = _load_ftr_mean_std(hparams.std_file)
        hparams.add_hparam('ftr_mean', np.array(ftr_mean, dtype=np.float32))
        hparams.add_hparam('ftr_std', np.array(ftr_std, dtype=np.float32))

    if hparams.explicit_empty:
        assert hparams.ftr_ext == 'cnn', 'explicit_empty will only be True when ftr_ext is cnn'

    # Checking hparam keep_checkpoint_max: must be >= 0
    if hparams.keep_checkpoint_max:
        assert hparams.keep_checkpoint_max >= 0

    # Classification task
    if hparams.num_classes > 1:
        # For classification tasks, restrict pmetric to be accuracy and use accuracy and confusion_matrix as metrics.
        hparams.pmetric = 'accuracy'
        hparams.all_metrics = ['accuracy', 'confusion_matrix']

    # L1 and L2 scale must be non-negative values
    assert hparams.l1 is None or hparams.l1 >= 0, "l1 scale must be non-negative"
    assert hparams.l2 is None or hparams.l2 >= 0, "l1 scale must be non-negative"

    # Multi-task training: currently only support ranking tasks with both deep and wide features
    if hparams.task_ids:
        # Check related inputs for multi-task training
        assert 'wide_ftrs_sp_idx' not in hparams.feature_names, "multi-task with sparse features not supported"
        assert 'task_id' in hparams.feature_names, "task_id feature not found for multi-task training"

        # Parse task ids an weights from inputs and convert them into a map
        task_ids = hparams.task_ids
        raw_weights = hparams.task_weights if hparams.task_weights else [1.0] * len(task_ids)
        task_weights = [float(wt) / sum(raw_weights) for wt in raw_weights]  # Normalize task weights

        # Check size of task_ids and task_weights
        assert len(task_ids) == len(task_weights), "size of task IDs and weights must match"

        force_set_hparam(hparams, "task_weights", task_weights)

    return hparams


def clean_tfreocrds(input_file, output_file):
    """Clean tfrecords file"""
    writer = tf.python_io.TFRecordWriter(output_file)
    count = 0
    for example in tf.python_io.tf_record_iterator(input_file):
        result = tf.train.Example.FromString(example)
        labels = result.features.feature['label'].float_list.value
        # the labels should have at least 2 different values
        if len(set(labels)) != 1:
            writer.write(example)
        else:
            print(labels)
            count += 1
    print(str(count) + ' examples has the same labels')
    writer.close()


def shuffle_tfrecords(input_file, output_file):
    """shuffle tfrecords file"""
    writer = tf.python_io.TFRecordWriter(output_file)
    data = []
    for example in tf.python_io.tf_record_iterator(input_file):
        data.append(example)

    random.shuffle(data)
    for example in data:
        writer.write(example)
    writer.close()


def sample_tfrecords(input_file, sample_cnt, output_file):
    """sample tfrecords file"""
    writer = tf.python_io.TFRecordWriter(output_file)
    data = []
    for example in tf.python_io.tf_record_iterator(input_file):
        data.append(example)

    random.shuffle(data)
    cnt = 0
    for example in data:
        writer.write(example)
        cnt += 1
        if cnt == sample_cnt:
            break
    writer.close()


def random_baseline(input_files, topk):
    """compute random baseline NDCG"""
    if type(input_files) is not list:
        input_files = [input_files]
    ndcg_scores = []
    count = 0
    for input_file in input_files:
        for example in tf.python_io.tf_record_iterator(input_file):
            count += 1
            result = tf.train.Example.FromString(example)
            label = result.features.feature['label'].float_list.value
            for _ in range(1):
                scores = random.sample(range(len(label)), len(label))
                ndcg_scores.append(test_utils.get_ndcg(scores, label, topk))
    print(count)
    print(f"Random baseline NDCG : {np.mean(ndcg_scores)}")


def generate_latency_test_data(input_file, output_file, field_names, target_docs, num_wide):
    """
    Generate the data for latency test.
    For example, one query has 100 documents.
    """
    count = 0
    records = []
    # read data
    for example in tf.python_io.tf_record_iterator(input_file):
        result = tf.train.Example.FromString(example)
        records.append(result)
        count += 1
        # at most 1000 queries
        if count == 1000:
            break
    print(f"read {count}", "records")

    # for each query
    with open(output_file, 'w') as fout:
        for i in range(len(records)):
            # query
            query = records[i].features.feature['query'].bytes_list.value[0].decode('utf-8')
            fout.write(query)
            fields = [[] for _ in field_names]
            j = 0
            while j < target_docs:
                index = random.randint(0, len(records) - 1)
                result = records[index]
                for k, field_name in enumerate(field_names):
                    if field_name == 'wide_ftrs':
                        ftrs = result.features.feature['wide_ftrs'].float_list.value
                    else:
                        ftrs = result.features.feature[field_name].bytes_list.value
                        ftrs = [x.decode('utf-8') for x in ftrs]
                    fields[k].extend(ftrs)
                j += len(result.features.feature['wide_ftrs'].float_list.value) / num_wide
            # wide features
            wide_ftrs_str = ' '.join(str(x) for x in fields[0][:target_docs * num_wide])
            fout.write('\t' + wide_ftrs_str)
            for field in fields[1:]:
                field_text = '**weiweiguo**'.join(field[:target_docs])
                fout.write('\t' + field_text)
            fout.write('\n')


def data_stats(input_files):
    """compute data statistics, such as click per result"""
    if type(input_files) is not list:
        input_files = [input_files]
    click_per_search = []
    for input_file in input_files:
        for example in tf.python_io.tf_record_iterator(input_file):
            result = tf.train.Example.FromString(example)
            label = result.features.feature['label'].float_list.value
            click_per_search.append(np.mean(label))
    print("{} : {}".format("Clicks per result", np.mean(click_per_search)))


def generate_unit_test_query_embeddings(input_file, savedmodel_dir, output_file, num_samples):
    """
    Generating query embeddings for online unit test
    :param input_file: tfrecord data containing query field
    :param savedmodel_dir: savedmodel path
    :param output_file: file name to embeddings
    :param num_samples: limit number of sample from input_file to write
    :return:
    """
    with tf.Session(graph=tf.Graph()) as sess:
        tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], savedmodel_dir)
        query_placeholder = tf.get_default_graph().get_tensor_by_name('query_placeholder:0')
        query_embedding = tf.get_default_graph().get_tensor_by_name('query_ftrs:0')

        query_list = []
        query_embedding_list = []
        for example in tf.python_io.tf_record_iterator(input_file):
            result = tf.train.Example.FromString(example)
            query = result.features.feature['query'].bytes_list.value[0].decode('utf-8')
            query_list.append(query)
            query_embeddings_v = sess.run(
                query_embedding,
                feed_dict={
                    query_placeholder: [query],
                }
            )
            query_embedding_list.append(query_embeddings_v[0])
            if len(query_embedding_list) == num_samples:
                break

        with open(output_file, 'w') as fout:
            for q, qe in zip(query_list, query_embedding_list):
                fout.write(q + ',' + ','.join(str(e) for e in qe))
                fout.write('\n')


def get_input_files(input_patterns):
    """Returns a list of file paths that match every pattern in input_patterns

    :param input_patterns a comma-separated string
    :return list of file paths
    """
    input_files = []
    for input_pattern in input_patterns.split(","):
        if tf.io.gfile.isdir(input_pattern):
            input_pattern = os.path.join(input_pattern, '*')
        input_files.extend(tf.gfile.Glob(input_pattern))
    return input_files


def _load_ftr_mean_std(path):
    """ Loads mean and standard deviation from given file """
    with tf.gfile.Open(path, 'rb') as fin:
        if path.endswith("fromspark"):
            data = fin.readlines()
            # Line 0 is printing message, line 1 is feature mean, line 2 is feature std
            ftr_mean = [float(x.strip()) for x in data[1].decode("utf-8").split(',')]
            ftr_std = [float(x.strip()) for x in data[2].decode("utf-8").split(',')]
        else:
            ftr_mean, ftr_std = pickle.load(fin)
    # Put std val 0 -> 1 to avoid zero division error
    for i in range(len(ftr_std)):
        if ftr_std[i] == 0:
            ftr_std[i] = 1
    return ftr_mean, ftr_std


def estimate_train_steps(input_pattern, num_epochs, batch_size, isTfrecord):
    """
    Estimate train steps from number of epochs.
    Counting exact total nubmer of examples is time consuming and unnecessary,
    we count the first file and use the total file size to estimate
    total nubmer of examples.
    """
    # TODO: for now throw error if input file is avro format
    estimated_num_examples = 0
    tf.logging.info("Estimating train steps of {} epochs".format(num_epochs))
    if not isTfrecord:
        raise ValueError("--num_epochs doesn't support avro yet.")
    else:
        input_files = get_input_files(input_pattern)

        file_1st = input_files[0]
        file_1st_num_examples = sum(1 for _ in tf.python_io.tf_record_iterator(file_1st))
        tf.logging.info("number of examples in first file: {0}".format(file_1st_num_examples))

        file_1st_size = tf.gfile.GFile(file_1st).size()
        tf.logging.info("first file size: {0}".format(file_1st_size))

        file_size_num_example_ratio = float(file_1st_size) / file_1st_num_examples

        estimated_num_examples = sum([int(tf.gfile.GFile(fn).size() / file_size_num_example_ratio)
                                      for fn in input_files])

    tf.logging.info("Estimated number of examples: {0}".format(estimated_num_examples))

    num_train_steps = int(estimated_num_examples * num_epochs / batch_size)
    tf.logging.info("{0} epochs approximately need {1} train steps with batch size {2}".format(
        num_epochs,
        num_train_steps,
        batch_size))

    return num_train_steps



import tensorflow as tf
import numpy as np
import pickle
import sys
import random
from avro.datafile import DataFileReader
from avro.io import DatumReader


MAX_FILE = 50
random.seed(4321)


def get_mean_std(input_files_patten, norm_file, num_wide):
    """Compute mean and std"""
    file_format = input_files_patten.split('.')[-1]
    all_ftrs = [[] for _ in range(num_wide)]
    tf.logging.set_verbosity(tf.logging.INFO)
    # read data, apply normalization on wide features
    files = tf.gfile.Glob(input_files_patten)
    random.shuffle(files)
    if file_format == 'tfrecords':
        for input_file in files[:MAX_FILE]:
            tf.logging.info(input_file)
            for example in tf.python_io.tf_record_iterator(input_file):
                result = tf.train.Example.FromString(example)
                wide_ftrs = result.features.feature['wide_ftrs'].float_list.value
                for i, f in enumerate(wide_ftrs):
                    i2 = i % num_wide
                    all_ftrs[i2].append(f)
    elif file_format == 'avro':
        for input_file in files[:MAX_FILE]:
            tf.logging.info(input_file)
            reader = DataFileReader(tf.gfile.Open(input_file, "rb"), DatumReader())
            for record in reader:
                wide_ftrs = record['wide_ftrs']
                for i, f in enumerate(wide_ftrs):
                    i2 = i % num_wide
                    all_ftrs[i2].append(f)
    else:
        raise ValueError("Unsupported file format %s." % file_format)
    mean_value = [0] * num_wide
    std_value = [0] * num_wide
    for i in range(num_wide):
        tf.logging.info('%d %f %f' % (i, np.mean(all_ftrs[i]), np.std(all_ftrs[i])))
        mean_value[i] = np.mean(all_ftrs[i]).item()
        std_value[i] = np.std(all_ftrs[i]).item()
    with tf.gfile.Open(norm_file, 'wb') as fout:
        pickle.dump((mean_value, std_value), fout, protocol=2)


if __name__ == '__main__':
    if len(sys.argv) != 4:
        print('Usage:  python  get_mean_std.py  input_files_patten  output_file  num_wide')
        sys.exit()
    input_files_patten = sys.argv[1]
    output_file = sys.argv[2]
    num_wide = int(sys.argv[3])
    get_mean_std(input_files_patten, output_file, num_wide)



import tensorflow as tf


class SparseEmbModel:
    """Embedding model that performs embedding lookup and summation on sparse features"""

    def __init__(self,
                 num_wide_sp: int,
                 wide_ftrs_sp_idx: tf.Tensor,
                 sp_emb_size: int,
                 wide_ftrs_sp_val: tf.Tensor = None,
                 padding_idx: int = 0,
                 initializer=tf.contrib.layers.xavier_initializer()):
        """ Computes a embedding given wide feature indices and values
        If wide_ftrs_sp_val is specified, users should keep consistency between wide_ftrs_sp_idx and wide_ftrs_sp_val
            -- the value of wide_ftrs_sp_idx[i] should be wide_ftrs_sp_val[i].
        CAVEAT: it is required that padding value = 0 for wide_ftrs_sp_idx in data_fn for
            wide_ftrs_sp_val since feature index 0 is used as the bias term.

        :param wide_ftrs_sp_idx: Feature indices. Shape=[batch_size, max_group_size, max_wide_ftrs_size]
        :param wide_ftrs_sp_val: Feature values. Shape=[batch_size, max_group_size, max_wide_ftrs_size]
        :param num_wide_sp: Number of unique sparse wide features. This should be LARGER than the maximum of
            features indices. Remember to add at least 1 to max(indices) avoid overflow due to padding
        :param sp_emb_size: Embedding size of sparse features
        :param padding_idx: Indices of padding in wide_ftrs_sp_idx. Recommend not to change it
        :param initializer: Weight initializer
        """
        wide_ftrs_sp_idx = tf.cast(wide_ftrs_sp_idx, dtype=tf.float32)
        if wide_ftrs_sp_val is None:  # Default to 1 if values unspecified
            wide_ftrs_sp_val = tf.ones(tf.shape(wide_ftrs_sp_idx), dtype=tf.float32)

        self._num_wide_sp = num_wide_sp
        self._padding_idx = padding_idx

        with tf.variable_scope('wide', reuse=tf.AUTO_REUSE):
            # Feature weights
            self.ftrs_weight = tf.get_variable('wide_ftrs_sp_weight',
                                               shape=[num_wide_sp, sp_emb_size],
                                               initializer=initializer,
                                               trainable=True)

            # A hack to combine idx and val so that we can process them together in `tf.map_fn` later
            # Shape=[batch_size, max_group_size, max_wide_ftrs_sp_size*2], max_wide_ftrs_size is the maximum number of
            #  sparse wide features in a document in the batch
            wide_ftrs_sp_idx_with_value = tf.concat([wide_ftrs_sp_idx, wide_ftrs_sp_val], axis=-1)

            # Compute embedding sample-wise
            self.embedding = tf.map_fn(self._compute_embedding_score_per_record, wide_ftrs_sp_idx_with_value,
                                       dtype=tf.float32)

    def _compute_embedding_score_per_record(self, wide_ftrs_idx_with_value_per_record):
        """Computes the embedding for given wide_ftrs_sp_idx and wide_ftrs_sp_val.
        A SparseTensor is created using wide_ftrs_sp_idx and wide_ftrs_sp_val and then multiplied with weights to
            obtain an embedding for each document
        """

        # Split idx and val back
        wide_ftrs_sp_idx, wide_ftrs_sp_val = tf.split(wide_ftrs_idx_with_value_per_record, 2, axis=-1)
        wide_ftrs_sp_idx = tf.cast(wide_ftrs_sp_idx, dtype=tf.int64)

        # Transformation
        shape = tf.ones(shape=[tf.shape(wide_ftrs_sp_idx)[0], 1], dtype=tf.int64) * self._num_wide_sp
        valid_wide_ftrs_idx_mask = tf.cast(tf.not_equal(wide_ftrs_sp_idx, self._padding_idx), tf.float32)
        wide_ftrs_sp_idx = tf.expand_dims(wide_ftrs_sp_idx, -1)

        # Get sparse feature vector v where v[ftr_idx_i] = ftr_val_i and v[other] = 0
        wide_ftrs_sp = sparse_tensor_merge(wide_ftrs_sp_idx,
                                           wide_ftrs_sp_val * valid_wide_ftrs_idx_mask,
                                           shape)

        # Feature weights
        bias = self.ftrs_weight[0]

        # Compute embedding
        embedding = tf.sparse.matmul(wide_ftrs_sp, self.ftrs_weight) + bias
        return embedding


def sparse_tensor_merge(indices: tf.Tensor, values: tf.Tensor, shape: tf.Tensor) -> tf.SparseTensor:
    """Creates a SparseTensor from grouped indices, values, and shapes
    Copied from https://stackoverflow.com/questions/42147362/create-sparsetensor-from-dequeued-batch-of-values-indices-shapes

    Args:
      indices: A [batch_size, N, D] integer Tensor
      values: A [batch_size, N] Tensor of any dtype
      shape: A [batch_size, D] Integer Tensor
    Returns:
      A SparseTensor of dimension D + 1 with batch_size as its first dimension.
    """
    indices = tf.cast(indices, dtype=tf.int64)
    shape = tf.cast(shape, dtype=tf.int64)
    merged_shape = tf.reduce_max(shape, axis=0)
    batch_size, elements, shape_dim = tf.unstack(tf.shape(indices))
    index_range_tiled = tf.tile(tf.range(batch_size)[..., None],
                                tf.stack([1, elements]))[..., None]
    merged_indices = tf.reshape(
        tf.concat([tf.cast(index_range_tiled, tf.int64), indices], axis=2),
        [-1, 1 + tf.size(merged_shape)])
    merged_values = tf.reshape(values, [-1])
    concat_shape = tf.concat([[tf.cast(batch_size, tf.int64)], merged_shape], axis=0)

    return tf.SparseTensor(merged_indices, merged_values, concat_shape)



import tensorflow as tf


class Logger(object):
    """
    Logs the printed messages also to a log file. This redirect the printed info to stderr (for faster print out)
    and the log_file
    """

    def __init__(self, log_file):
        """

        :param log_file: The log_file filename for directing the messages to. Could be a local file or an hdfs file
        """
        tf.logging.set_verbosity(tf.logging.INFO)
        self.tf_logging = tf.logging
        self.writer = tf.gfile.Open(log_file, 'w')

    def write(self, message):
        """
        Writes to stderr, and log_file

        :param message: The message to write
        :return:
        """
        if message and message != '\n' and (not message.isspace()):
            self.tf_logging.info(message)
        self.writer.write(message)

    def flush(self):
        """

        This flush method is needed for python 3 compatibility.
        This handles the flush command by doing nothing.
        :return:
        """
        pass



"""
Overall pipeline to train the model.  It parses arguments, and trains a DeText model.
"""

import logging
import os
import sys
import time
from typing import NamedTuple, List, Optional, Dict

import tensorflow as tf
import tensorflow_ranking as tfr
from detext.train import train
from detext.train.data_fn import input_fn
from detext.utils import misc_utils, logger
from detext.utils.executor_utils import EVALUATOR, LOCAL_MODE, CHIEF, get_executor_task_type
from linkedin.smart_arg import arg_suite
from tensorflow.contrib.training import HParams


@arg_suite
class DetextArg(NamedTuple):
    """
    DeText: a Deep Text understanding framework for NLP related ranking, classification, and language generation tasks.
    It leverages semantic matching using deep neural networks to
    understand member intents in search and recommender systems.
    As a general NLP framework, currently DeText can be applied to many tasks,
    including search & recommendation ranking, multi-class classification and query understanding tasks.
    """
    #  kwargs utility function for split a str to a List, provided for backward compatibilities
    #  Please use built-in List parsing support when adding new arguments
    __comma_split_list = lambda t: {'type': lambda s: [t(item) for item in s.split(',')] if ',' in s else t(s), 'nargs': None}  # noqa: E731

    feature_names: List[str]  # Feature names.

    # network
    ftr_ext: str  # NLP feature extraction module.
    _ftr_ext = {'choices': ['cnn', 'bert', 'lstm_lm', 'lstm']}
    num_units: int = 128  # word embedding size.
    num_units_for_id_ftr: int = 128  # id feature embedding size.
    sp_emb_size: int = 1  # Embedding size of sparse features
    num_hidden: List[int] = [0]  # hidden size.
    num_wide: int = 0  # number of wide features per doc.
    num_wide_sp: Optional[int] = None  # number of sparse wide features per doc
    use_deep: bool = True  # Whether to use deep features.
    _use_deep = {'type': lambda s: True if s.lower() == 'true' else False if s.lower() == 'false' else s}
    elem_rescale: bool = True  # Whether to perform elementwise rescaling.
    use_doc_projection: bool = False  # whether to project multiple doc features to 1 vector.
    use_usr_projection: bool = False  # whether to project multiple usr features to 1 vector.

    # Ranking specific
    ltr_loss_fn: str = 'pairwise'  # learning-to-rank method.
    emb_sim_func: List[str] = ['inner']  # Approach to computing query/doc similarity scores
    _emb_sim_func = {"choices": ('inner', 'hadamard', 'concat', 'diff')}

    # Classification specific
    num_classes: int = 1  # Number of classes for multi-class classification tasks.

    # CNN related
    filter_window_sizes: List[int] = [3]  # CNN filter window sizes.
    num_filters: int = 100  # number of CNN filters.
    explicit_empty: bool = False  # Explicitly modeling empty string in cnn

    # BERT related
    lr_bert: Optional[float] = None  # Learning rate factor for bert components
    bert_config_file: Optional[str] = None  # bert config.
    bert_checkpoint: Optional[str] = None  # pretrained bert model checkpoint.
    use_bert_dropout: bool = False  # apply dropout during training in bert layers.

    # LSTM related
    unit_type: str = 'lstm'  # RNN cell unit type. Currently only supports lstm. Will support other cell types in the future
    _unit_type = {'choices': ['lstm']}
    num_layers: int = 1  # RNN layers
    num_residual_layers: int = 0  # Number of residual layers from top to bottom. For example, if `num_layers=4` and `num_residual_layers=2`, the last 2 RNN cells in the returned list will be wrapped with `ResidualWrapper`. # noqa: E501
    forget_bias: float = 1.  # Forget bias of RNN cell
    rnn_dropout: float = 0.  # Dropout of RNN cell
    bidirectional: bool = False  # Whether to use bidirectional RNN
    normalized_lm: bool = False  # Whether to use normalized lm. This option only works for unit_type=lstm_lm

    # Optimizer
    optimizer: str = 'sgd'  # Type of optimizer to use. bert_adam is similar to the optimizer implementation in bert.
    _optimizer = {'choices': ['sgd', 'adam', 'bert_adam', 'bert_lamb']}
    max_gradient_norm: float = 1.0  # Clip gradients to this norm.
    learning_rate: float = 1.0  # Learning rate. Adam: 0.001 | 0.0001
    num_train_steps: int = 1  # Num steps to train.
    num_epochs: Optional[int] = None  # Num of epochs to train, will overwrite train_steps if set
    num_warmup_steps: int = 0  # Num steps for warmup.
    train_batch_size: int = 32  # Training data batch size.
    test_batch_size: int = 32  # Test data batch size.
    l1: Optional[float] = None  # Scale of L1 regularization
    l2: Optional[float] = None  # Scale of L2 regularization

    # Data
    train_file: Optional[str] = None  # Train file.
    dev_file: Optional[str] = None  # Dev file.
    test_file: Optional[str] = None  # Test file.
    out_dir: Optional[str] = None  # Store log/model files.
    std_file: Optional[str] = None  # feature standardization file
    max_len: int = 32  # max sent length.
    min_len: int = 3  # min sent length.

    # Vocab and word embedding
    vocab_file: Optional[str] = None  # Vocab file
    we_file: Optional[str] = None  # Pretrained word embedding file
    we_trainable: bool = True  # Whether to train word embedding
    PAD: str = '[PAD]'  # Token for padding
    SEP: str = '[SEP]'  # Token for sentence separation
    CLS: str = '[CLS]'  # Token for start of sentence
    UNK: str = '[UNK]'  # Token for unknown word
    MASK: str = '[MASK]'  # Token for masked word

    # Vocab and word embedding for id features
    vocab_file_for_id_ftr: Optional[str] = None  # Vocab file for id features
    we_file_for_id_ftr: Optional[str] = None  # Pretrained word embedding file for id features
    we_trainable_for_id_ftr: bool = True  # Whether to train word embedding for id features
    PAD_FOR_ID_FTR: str = '[PAD]'  # Padding token for id features
    UNK_FOR_ID_FTR: str = '[UNK]'  # Unknown word token for id features

    # Misc
    random_seed: int = 1234  # Random seed (>0, set a specific seed).
    steps_per_stats: int = 100  # training steps to print statistics.
    num_eval_rounds: Optional[int] = None  # number of evaluation round, this param will override steps_per_eval as max(1, num_train_steps / num_eval_rounds)
    steps_per_eval: int = 1000  # training steps to evaluate datasets.
    keep_checkpoint_max: int = 5  # The maximum number of recent checkpoint files to keep. If 0, all checkpoint files are kept. Defaults to 5
    _feature_names = __comma_split_list(str)
    lambda_metric: Optional[str] = None  # only support ndcg.
    init_weight: float = 0.1  # weight initialization value.
    pmetric: Optional[str] = None  # Primary metric.
    all_metrics: Optional[List[str]] = None  # All metrics.
    score_rescale: Optional[List[float]] = None  # The mean and std of previous model. For score rescaling, the score_rescale has the xgboost mean and std.

    add_first_dim_for_query_placeholder: bool = False  # Whether to add a batch dimension for query and usr_* placeholders. This shall be set to True if the query field is used as document feature in model serving.# noqa: E501
    add_first_dim_for_usr_placeholder: bool = False  # Whether to add a batch dimension for query and usr_* placeholders. This shall be set to True if usr fields are used document feature in model serving.# noqa: E501

    tokenization: str = 'punct'  # The tokenzation performed for data preprocessing. Currently support: punct/plain(no split). Note that this should be set correctly to ensure consistency for savedmodel.# noqa: E501
    _tokenization = {'choices': ['plain', 'punct']}

    resume_training: bool = False  # Whether to resume training from checkpoint in out_dir.
    metadata_path: Optional[str] = None  # The metadata_path for converted avro2tf avro data.

    # tf-ranking related
    use_tfr_loss: bool = False  # whether to use tf-ranking loss.
    tfr_loss_fn: str = tfr.losses.RankingLossKey.SOFTMAX_LOSS  # tf-ranking loss
    _tfr_loss_fn = {'choices': [tfr.losses.RankingLossKey.SOFTMAX_LOSS, tfr.losses.RankingLossKey.PAIRWISE_LOGISTIC_LOSS]}
    tfr_lambda_weights: Optional[str] = None  #

    use_horovod: bool = False  # whether to use horovod for sync distributed training
    hvd_info: Optional[Dict[str, int]] = None
    # multitask training related
    task_ids: Optional[List[int]] = None  # All types of task IDs for multitask training. E.g., 1,2,3
    _task_ids = __comma_split_list(int)
    task_weights: Optional[List[float]] = None  # Weights for each task specified in task_ids. E.g., 0.5,0.3,0.2
    _task_weights = __comma_split_list(float)

    # This method is automatically called by smart-arg once the argument is created by parsing cli or the constructor
    # It's used to late-initialize some fields after other fields are created.
    def __late_init__(self):
        arg = self
        # if epoch is set, overwrite training steps
        if arg.num_epochs is not None:
            arg = arg._replace(num_train_steps=misc_utils.estimate_train_steps(
                arg.train_file,
                arg.num_epochs,
                arg.train_batch_size,
                arg.metadata_path is None))

        # if num_eval_rounds is set, override steps_per_eval
        if arg.num_eval_rounds is not None:
            arg = arg._replace(steps_per_eval=max(1, int(arg.num_train_steps / arg.num_eval_rounds)))

        # For data sharding when using horovod
        if arg.use_horovod:
            import horovod.tensorflow as hvd
            hvd.init()
            arg = arg._replace(
                num_train_steps=arg.num_train_steps // hvd.size(),
                num_warmup_steps=arg.num_warmup_steps // hvd.size(),
                steps_per_eval=arg.steps_per_eval // hvd.size(),
                steps_per_stats=arg.steps_per_stats // hvd.size(),
                hvd_info={'rank': hvd.rank(), 'size': hvd.size()})

        return arg


def get_hparams(argument: DetextArg = None):
    """
    Get hyper-parameters.
    Not in use and to be removed.
    """
    import warnings
    warnings.warn("get_hparams is deprecated and not used, call `run_detext` directly")
    hparams = HParams(**(argument or DetextArg.__from_argv__(error_on_unknown=False))._asdict())
    # Print all hyper-parameters
    for k, v in sorted(vars(hparams).items()):
        logging.info(f'{k} = {v}')

    return hparams


def main(argv):
    """
    This is the main method for training the model.
    :param argv: training parameters
    """

    argument = DetextArg.__from_argv__(argv[1:], error_on_unknown=False)
    run_detext(argument)


def run_detext(argument: DetextArg):
    logging.info(f"Args:\n {argument}")

    tf.logging.set_verbosity(tf.logging.INFO)

    # For data sharding when using horovod
    if argument.use_horovod:
        import horovod.tensorflow as hvd
    else:
        hvd = None

    # Get executor task type from TF_CONFIG
    task_type = get_executor_task_type()

    # Create directory and launch tensorboard
    master = not hvd or hvd.rank() == 0
    if (task_type == CHIEF or task_type == LOCAL_MODE) and master:
        # If not resume training from checkpoints, delete output directory.
        if not argument.resume_training:
            logging.info("Removing previous output directory...")
            if tf.gfile.Exists(argument.out_dir):
                tf.gfile.DeleteRecursively(argument.out_dir)

        # If output directory deleted or does not exist, create the directory.
        if not tf.gfile.Exists(argument.out_dir):
            logging.info('Creating dirs recursively at: {0}'.format(argument.out_dir))
            tf.gfile.MakeDirs(argument.out_dir)

        misc_utils.save_hparams(argument.out_dir, HParams(**argument._asdict()))

    else:
        # TODO: move removal/creation to a hadoopShellJob st. it does not reside in distributed training code.
        logging.info("Waiting for chief to remove/create directories.")
        # Wait for dir created form chief
        time.sleep(10)

    if (task_type == EVALUATOR or task_type == LOCAL_MODE) and master:
        # set up logger for evaluator
        sys.stdout = logger.Logger(os.path.join(argument.out_dir, 'eval_log.txt'))

    hparams = misc_utils.extend_hparams(HParams(**argument._asdict()))

    logging.info("***********DeText Training***********")

    # Train and evaluate DeText model
    train.train(hparams, input_fn)


if __name__ == '__main__':
    tf.compat.v1.app.run()



import pickle
import tensorflow as tf


def init_word_embedding(hparams, mode, name_prefix="w"):
    """Initialize word embeddings from random initialization or pretrained word embedding.

    This function is only used by encoding models other than BERT
    """

    we_trainable = hparams.get("we_trainable")
    we_file = hparams.get("we_file")
    vocab_size = hparams.get("vocab_size")
    num_units = hparams.get("num_units")

    if we_file is None:
        embedding_name = "{}_pretrained_embedding".format(name_prefix)
        # Random initialization
        embedding = tf.get_variable(
            embedding_name, [vocab_size, num_units], dtype=tf.float32, trainable=we_trainable)
    else:
        # Initialize by pretrained word embedding
        embedding_name = "{}_embedding".format(name_prefix)
        we = pickle.load(tf.gfile.Open(we_file, 'rb'))
        assert vocab_size == we.shape[0] and num_units == we.shape[1]
        embedding = tf.get_variable(
            name=embedding_name,
            shape=[vocab_size, num_units],
            dtype=tf.float32,
            initializer=tf.constant_initializer(we),
            trainable=we_trainable
        )
    return embedding


def _single_cell(unit_type, num_units, forget_bias, dropout,
                 mode, residual_connection=False, device_str=None):
    """Create an instance of a single RNN cell."""
    # dropout (= 1 - keep_prob) is set to 0 during eval
    dropout = dropout if mode == tf.contrib.learn.ModeKeys.TRAIN else 0.0
    # Cell Type
    if unit_type == "lstm":
        single_cell = tf.nn.rnn_cell.LSTMCell(num_units, forget_bias=forget_bias)
    elif unit_type == "gru":
        single_cell = tf.nn.rnn_cell.GRUCell(num_units)
    elif unit_type == "layer_norm_lstm":
        single_cell = tf.contrib.rnn.LayerNormBasicLSTMCell(
            num_units,
            forget_bias=forget_bias,
            layer_norm=True)
    else:
        raise ValueError("Unknown unit type %s!" % unit_type)

    # Dropout (= 1 - keep_prob)
    if dropout > 0.0:
        single_cell = tf.contrib.rnn.DropoutWrapper(cell=single_cell, input_keep_prob=(1.0 - dropout))

    # Residual
    if residual_connection:
        single_cell = tf.contrib.rnn.ResidualWrapper(single_cell)

    # Device Wrapper
    if device_str:
        single_cell = tf.contrib.rnn.DeviceWrapper(single_cell, device_str)

    return single_cell


def create_rnn_cell(unit_type, num_units, num_layers, num_residual_layers,
                    forget_bias, dropout, mode):
    """Create multi-layer RNN cell.

    Args:
      unit_type: string representing the unit type, i.e. "lstm".
      num_units: the depth of each unit.
      num_layers: number of cells.
      num_residual_layers: Number of residual layers from top to bottom. For
        example, if `num_layers=4` and `num_residual_layers=2`, the last 2 RNN
        cells in the returned list will be wrapped with `ResidualWrapper`.
      forget_bias: the initial forget bias of the RNNCell(s).
      dropout: floating point value between 0.0 and 1.0:
        the probability of dropout.  this is ignored if `mode != TRAIN`.
      mode: either tf.contrib.learn.TRAIN/EVAL
      num_gpus: The number of gpus to use when performing round-robin
        placement of layers.
      base_gpu: The gpu device id to use for the first RNN cell in the
        returned list. The i-th RNN cell will use `(base_gpu + i) % num_gpus`
        as its device id.
      single_cell_fn: allow for adding customized cell.
        When not specified, we default to _single_cell
    Returns:
      An `RNNCell` instance.
    """
    cell_list = []
    for i in range(num_layers):
        single_cell = _single_cell(
            unit_type=unit_type,
            num_units=num_units,
            forget_bias=forget_bias,
            dropout=dropout,
            mode=mode,
            residual_connection=(i >= num_layers - num_residual_layers),
        )
        cell_list.append(single_cell)

    if len(cell_list) == 1:  # Single layer.
        return cell_list[0]
    else:  # Multi layers
        return tf.contrib.rnn.MultiRNNCell(cell_list)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.
import setuptools
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Update (increment) the version number before releasing. Checkout published versions at
# https://pypi.org/manage/project/detext/releases/
# Please follow the following best practices for versioning: Breaking changes are indicated by increasing the
# major number (high risk), new non-breaking features increment the minor number (medium risk) and all other
# non-breaking changes increment the patch number (lowest risk).
VERSION = '2.0.8'

setuptools.setup(
    name='detext',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=["Programming Language :: Python :: 3.7",
                 "Intended Audience :: Science/Research",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved"],
    license='BSD-2-CLAUSE',
    version=VERSION,
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    include_package_data=True,
    install_requires=['numpy<1.17', 'smart-arg==0.0.5', 'tensorflow==1.14.0', 'tensorflow_ranking==0.1.4', 'gast==0.2.2'],
    tests_require=[
        'pytest',
    ])



"""
The deep match model.
"""
import tensorflow as tf

from detext.model import rep_model
from detext.model.sp_emb_model import SparseEmbModel


class DeepMatch:
    """
    The deep match model
    """

    def __init__(self,
                 query,
                 wide_ftrs,
                 doc_fields,
                 hparams,
                 mode,
                 wide_ftrs_sp_idx=None,
                 wide_ftrs_sp_val=None,
                 usr_fields=None,
                 doc_id_fields=None,
                 usr_id_fields=None,
                 task_id_field=None,
                 ):
        """
        Build the deep match graph.
        :param query: Tensor  Input query. Shape=[batch_size, query_length]
        :param wide_ftrs: Tensor  Wide features. Shape=[batch_size, max_group_size, num_wide_ftrs]
        :param wide_ftrs_sp_idx: Tensor  Sparse wide features indices. Shape=[batch_size, max_group_size, num_wide_ftrs]
        :param wide_ftrs_sp_val: Tensor  Sparse wide features values. Shape=[batch_size, max_group_size, num_wide_ftrs]
        :param doc_fields: list(Tensor) or a pre-computed Tensor  List of document fields. Each has
            shape=[batch_size, max_group_size, doc_field_length]
        :param usr_fields: list(Tensor) or a pre-computed Tensor  List of user fields. Each has
            shape=[batch_size, usr_field_length]
        :param doc_id_fields: list(Tensor)  List of user id fields. Each has shape=[batch_size, doc_id_field_length]
        :param usr_id_fields: list(Tensor)  List of doc id fields. Each has shape=[batch_size, usr_id_field_length]
        :param task_id_field: Tensor  Task id for multitask. Shape=[batch_size, ]
        :param hparams: HParams  Hyper parameters
        :param mode: TRAIN/EVAL/INFER
        """
        self._query = query
        self._wide_ftrs = tf.where(tf.is_nan(wide_ftrs), tf.zeros_like(wide_ftrs),
                                   wide_ftrs) if wide_ftrs is not None else wide_ftrs
        self._wide_ftrs_sp_idx = wide_ftrs_sp_idx
        self._wide_ftrs_sp_val = wide_ftrs_sp_val
        self._usr_fields = usr_fields
        self._doc_fields = doc_fields
        self._usr_id_fields = usr_id_fields
        self._doc_id_fields = doc_id_fields
        self._task_id_field = task_id_field
        self._hparams = hparams
        self._mode = mode

        self.batch_size, self.max_group_size = self.get_data_size()

        if hparams.use_deep is True:
            # Apply representation-based models
            self.deep_ftr_model = rep_model.RepModel(
                query=self._query,
                doc_fields=self._doc_fields,
                usr_fields=self._usr_fields,
                doc_id_fields=self._doc_id_fields,
                usr_id_fields=self._usr_id_fields,
                hparams=self._hparams,
                mode=self._mode
            )

        self.all_ftrs = self.compute_all_ftrs()
        self.scores = self.compute_scores(self.all_ftrs)
        self.original_scores = self.scores
        if self._mode == tf.estimator.ModeKeys.PREDICT and hparams.num_classes <= 1:
            # final_scores is transposed for ranking tasks
            self.scores = tf.transpose(self.scores, name='final_scores')

    def get_data_size(self):
        """Infers batch size, max group size"""
        if self._wide_ftrs is not None:
            data = self._wide_ftrs
        elif self._wide_ftrs_sp_idx is not None:
            data = self._wide_ftrs_sp_idx
        elif self._doc_fields is not None:
            data = self._doc_fields[0]
        elif self._doc_id_fields is not None:
            data = self._doc_id_fields[0]
        elif self._usr_fields is not None:
            data = self._usr_fields
        elif self._usr_id_fields:
            data = self._usr_id_fields
        else:
            raise ValueError('Cannot infer data size.')
        data_shape = tf.shape(data)
        return data_shape[0], data_shape[1]

    def compute_scores(self, all_ftrs):
        """Computes total scores """
        hparams = self._hparams

        # mlp and score layers
        if hparams.task_ids is not None:
            # derive individual task score shape
            score_shape = [self.batch_size, tf.maximum(self.max_group_size, hparams.num_classes)]
            scores = tf.zeros(shape=score_shape, dtype="float32")
            for task_id in hparams.task_ids:
                task_score = self.add_mlp_and_score_layers(
                    input_layer=all_ftrs,
                    prefix='task_' + str(task_id) + '_'
                )

                # use task_mask to zero out other tasks' score
                task_mask = tf.cast(tf.equal(self._task_id_field, int(task_id)), dtype=tf.float32)
                # broadcast task_mask for compatible tensor shape with scores tensor
                task_mask = tf.transpose(tf.broadcast_to(task_mask, score_shape[::-1]))
                scores += task_mask * task_score
        else:
            scores = self.add_mlp_and_score_layers(all_ftrs)

        return scores

    def compute_sp_ftrs(self):
        """Computes feature embedding from wide sparse features"""
        sp_emb_model = SparseEmbModel(wide_ftrs_sp_idx=self._wide_ftrs_sp_idx,
                                      num_wide_sp=self._hparams.num_wide_sp,
                                      wide_ftrs_sp_val=self._wide_ftrs_sp_val,
                                      sp_emb_size=self._hparams.sp_emb_size
                                      )
        return sp_emb_model, sp_emb_model.embedding

    def compute_all_ftrs(self):
        """
        Concatenates deep features, dense wide features and sparse wide feature embeddings
        """
        all_ftrs = []
        hparams = self._hparams
        if hparams.use_deep or self._wide_ftrs is not None:
            dense_ftrs = self.compute_dense_ftrs()
            all_ftrs.append(dense_ftrs)
        if self._wide_ftrs_sp_idx is not None:
            _, sp_embedding = self.compute_sp_ftrs()
            all_ftrs.append(sp_embedding)
        all_ftrs = tf.concat(all_ftrs, axis=-1)
        return all_ftrs

    def compute_dense_ftrs(self):
        """
        Computes the dense features based on dense features and deep features
        :return Tensor  Dense features. Shape=[batch_size, max_group_size, deep_ftr_size + dense_wide_ftr_size]
        """
        hparams = self._hparams
        batch_size = self.batch_size
        max_group_size = self.max_group_size

        num_sim_ftrs = hparams.num_doc_fields
        if hparams.use_deep:
            num_sim_ftrs, sim_ftrs = self.deep_ftr_model.num_sim_ftrs, self.deep_ftr_model.sim_ftrs
            # Assign a name to sim_ftrs in the graph
            sim_ftrs = tf.identity(sim_ftrs, name="sim_ftrs")

        if hparams.get('ftr_mean') is not None and self._wide_ftrs is not None:
            ftr_mean = tf.constant(hparams.ftr_mean, dtype=tf.float32)
            ftr_std = tf.constant(hparams.ftr_std, dtype=tf.float32)
            self._wide_ftrs = (self._wide_ftrs - ftr_mean) / ftr_std

        # elementwise rescaling for dense features
        if hparams.elem_rescale and self._wide_ftrs is not None:
            self.elem_norm_w = tf.get_variable("wide_ftr_norm_w", [hparams.num_wide], dtype=tf.float32,
                                               initializer=tf.constant_initializer(1.0))
            self.elem_norm_b = tf.get_variable("wide_ftr_norm_b", [hparams.num_wide], dtype=tf.float32,
                                               initializer=tf.constant_initializer(0.0))
            self._wide_ftrs = tf.tanh(self._wide_ftrs * self.elem_norm_w + self.elem_norm_b)

        emb_size = 0
        dense_ftrs = []

        # whether to include deep features
        if hparams.use_deep:
            dense_ftrs.append(sim_ftrs)
            emb_size += num_sim_ftrs

        # whether to include wide features
        if self._wide_ftrs is not None:
            dense_ftrs.append(self._wide_ftrs)
            emb_size += hparams.num_wide

        # add whether empty
        if hparams.explicit_empty:
            ftrs, size = self.add_empty_str_ftr(max_group_size)
            dense_ftrs += ftrs
            emb_size += size

        # reshape
        dense_ftrs = tf.concat(dense_ftrs, axis=-1)
        dense_ftrs = tf.reshape(dense_ftrs, shape=[batch_size, max_group_size, emb_size])
        return dense_ftrs

    def add_empty_str_ftr(self, max_group_size):
        """
        Add whether-the-string-is-empty features into dense_ftrs.
        """
        num_ftrs = 0
        all_ftrs = []
        hparams = self._hparams

        def process_text(text, is_doc=False):
            # text length
            text_length = tf.reduce_sum(tf.cast(tf.not_equal(text, hparams.pad_id), dtype=tf.int32), axis=-1)
            # whether the text is empty
            text_is_empty = tf.cast(tf.equal(text_length, max(hparams.filter_window_sizes) * 2 - 2), dtype=tf.float32)
            if not is_doc:
                # for non-doc, the shape is [batch_size]
                text_is_empty = tf.expand_dims(tf.expand_dims(text_is_empty, axis=-1), axis=-1)
                text_is_empty = tf.tile(text_is_empty, [1, max_group_size, 1])
            else:
                # for doc, the shape is [batch_size, max_group_size]
                text_is_empty = tf.expand_dims(text_is_empty, axis=-1)
            all_ftrs.append(text_is_empty)

        if self._query is not None:
            process_text(self._query)
            num_ftrs += 1
        if self._doc_fields is not None:
            for doc_field in self._doc_fields:
                process_text(doc_field, True)
                num_ftrs += 1
        if self._usr_fields is not None:
            for usr_field in self._usr_fields:
                process_text(usr_field)
                num_ftrs += 1
        return all_ftrs, num_ftrs

    def add_mlp_and_score_layers(self, input_layer, prefix=''):
        """
        Add multi-perception and scoring layers
        """
        hparams = self._hparams
        # hidden layer
        hidden_ftrs = input_layer
        for i, hidden_size in enumerate(hparams.num_hidden):
            if hidden_size == 0:
                continue
            hidden_ftrs = tf.layers.dense(
                hidden_ftrs,
                units=hidden_size,
                use_bias=True,
                activation=tf.tanh,
                name=f'{prefix}hidden_projection_{str(i)}')

        # final score for query/doc pairs
        scores = tf.layers.dense(
            hidden_ftrs,
            units=hparams.num_classes,
            use_bias=True,
            activation=None,
            name=f'{prefix}scoring')

        if hparams.num_classes <= 1:
            scores = tf.squeeze(scores, axis=-1)  # shape=[batch_size, max_group_size]
        else:
            scores = tf.squeeze(scores, axis=-2)  # shape=[batch_size, num_classes]

        return scores
